/* $Id: Socket.cc,v 1.18.2.7 1998/12/29 22:36:37 rousskov Exp $ */

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Socket.h"

#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h> /* for TCP_NODELAY */
#include <arpa/inet.h>
#include <netdb.h>
#include <assert.h>

#include "Bug.h"

bool Socket::IsEnabled = true;

bool Socket::create() {
	return create(PF_INET, SOCK_STREAM, 0);
}

bool Socket::create(int domain, int type, int protocol) {
	theFD = ::socket(domain, type, protocol);
	return theFD >= 0;
}

bool Socket::bind(int port) {
	static struct sockaddr_in S;

	S.sin_family = AF_INET;
	S.sin_port = htons(port);
	S.sin_addr.s_addr = htonl(INADDR_ANY);

	if (IsEnabled && ::bind(theFD, (struct sockaddr *) &S, sizeof(S)) != 0) {
		cerr << "bind to port " << port << " failed." << endl;
		if (errno != EADDRINUSE)
			cerr << bug;
		else
			cerr << endl << "cause(?): Address already in use" << endl;
		return false;
	}
	return true;
}

bool Socket::connect(const char *host, int port) {

	struct sockaddr_in s;
	memset(&s, 0, sizeof(s));

	s.sin_family = AF_INET;
	s.sin_addr.s_addr = inet_addr(host);
	s.sin_port = htons(port);

	if (IsEnabled && ::connect(theFD, (struct sockaddr *) &s, sizeof(s)) < 0) {
		if (errno != EINPROGRESS) { // XXX: || isBlocking()) {
			cerr << here << "fd: " << theFD << " connect from port " << this->lport() << " to " << host << ':' << port << " failed";
			if (errno != EADDRINUSE)
				cerr << bug;
			else
				cerr << endl << "cause(?): Address already in use; port: " << this->lport() << endl;
			return false;
		}
	}

	return true;
}

bool Socket::listen(int port) {

	if (!bind(port))
		cerr << "bind failed" << bug;

	if (IsEnabled && ::listen(theFD, FD_SETSIZE) < 0)
		cerr << "listen failed" << bug;

	return true;
}

Socket Socket::accept(struct sockaddr *addr_p, int *addr_len_p) {
	Socket s;
	struct sockaddr addr;
	int addr_len = sizeof(addr);
	if (!addr_p) addr_p = &addr;
	if (!addr_len_p) addr_len_p = &addr_len;
	s.theFD = IsEnabled ? ::accept(theFD, addr_p, addr_len_p) : (errno = EWOULDBLOCK, -1);
	return s;
}

ssize_t Socket::read(void *buf, size_t sz) {
	return IsEnabled ? ::read(theFD, buf, sz) : 0;
}

ssize_t Socket::write(const void *buf, size_t sz) {
	return IsEnabled ? ::write(theFD, buf, sz) : sz;
}

bool Socket::close() {

	const int res = ::close(theFD);
	theFD = -1;
	return res == 0;
}

void Socket::sendBuf(int sz) {
	assert(setOpt(SO_SNDBUF, (const void*)&sz, sizeof(sz)));
}

void Socket::recvBuf(int sz) {
	assert(setOpt(SO_RCVBUF, (const void*)&sz, sizeof(sz)));
}

int Socket::flags() const {

	int flags = fcntl(theFD, F_GETFL);

	if (flags < 0)
		cerr << here << "get socket flags failed" << bug;

	return flags;
}

/* local port */
int Socket::lport() const {
	struct sockaddr_in S;
	int len = sizeof(S);

	if (getsockname(theFD, (struct sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

/* remote (peer) port */
int Socket::rport() const {
	struct sockaddr_in S;
	int len = sizeof(S);

	if (getpeername(theFD, (struct sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

timeval Socket::rcvTimeout() const {
	struct timeval t;
	int len = sizeof(t);
	if (getOpt(SO_RCVTIMEO, (void*)&t, &len))
		return t;
	const timeval badt = { -1, -1 };
	return badt;
}

int Socket::sendBuf() const {
	int sz;
	int len = sizeof(sz);
	if (getOpt(SO_SNDBUF, (void*)&sz, &len))
		return sz;
	else
		return -1;
}

int Socket::recvBuf() const {
	int sz;
	int len = sizeof(sz);
	if (getOpt(SO_RCVBUF, (void*)&sz, &len))
		return sz;
	else
		return -1;
}

int Socket::errNo() const {
	int e;
	int len = sizeof(e);

	if (getOpt(SO_ERROR, (void*)&e, &len))
		return e;
	else
		return -1;
}

bool Socket::peerAddr(struct sockaddr *S, int &len) const {
	return getpeername(theFD, S, &len) == 0;
}

Address Socket::peerAddr() const {
	struct sockaddr_in s;
	int len = sizeof(s);
	if (peerAddr((struct sockaddr *) &s, len))
		return Address(inet_ntoa(s.sin_addr), ntohs(s.sin_port));
	else
		return Address("??", -1);
}

void Socket::reusePort(bool set) {
#ifdef SO_REUSEPORT
	setOpt(SO_REUSEPORT, set);
#else
	set = false; // make compiler happy
	// XXX: but no warning?
	#warning SO_REUSEPORT is not supported on this system
#endif
}

void Socket::linger(int delay) {
#ifdef SO_LINGER
	static struct linger l = { delay >= 0, delay };
	setOpt(SO_LINGER, &l, sizeof(l));
#else
	delay = 0; // make compiler happy
	// XXX: but no run-time warning?
	#warning SO_LINGER is not supported on this system
#endif
}

void Socket::blocking(bool set) {

	int flags = this->flags();

	if (set)
		flags &= ~O_NONBLOCK;
	else
		flags |= O_NONBLOCK;

	if (fcntl(theFD, F_SETFL, flags) == -1)
		cerr << here << "set socket flags failed" << bug;
}

void Socket::tcpDelay(bool set) {
#ifdef TCP_NODELAY
	int optVal = (int)!set;
    if (::setsockopt(theFD, IPPROTO_TCP, TCP_NODELAY, (void*)&optVal, sizeof(optVal)) != 0)
        cerr << here << "fd " << theFD << " failed to set TCP_NODELAY" << bug;
#else
	set = false; // make compiler happy
	#warning TCP_NODELAY is not supported on this system
#endif
}

bool Socket::isBlocking() const {
	return flags() & O_NONBLOCK == 0;
}


bool Socket::setOpt(int optName, const void *optVal, int optLen) {
	// XXX: maybe assert?
	if (::setsockopt(theFD, SOL_SOCKET, optName, optVal, optLen) != 0) {
		cerr << here << "fd " << theFD << " failed to change opt " << optName << " cause(?): " << strerror(errno) << endl;
		return false;
	}
	return true;
}

bool Socket::getOpt(int optName, void *optVal, int *optLen) const {
	// XXX: maybe assert?
	if (::getsockopt(theFD, SOL_SOCKET, optName, optVal, optLen) != 0) {
		cerr << here << "fd " << theFD << " failed to get opt " << optName << " cause(?): " << strerror(errno) << endl;
		return false;
	}
	return true;
}

bool Socket::setOpt(int optName, bool set) {
	int optVal = (int)set;
	return setOpt(optName, &optVal, sizeof(optVal));
}
