#ifndef _MSG_EVENT_H_
#define _MSG_EVENT_H_ "$Id: MsgEvent.h,v 1.6.2.1 1998/09/22 07:44:13 rousskov Exp $"

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Event.h"
#include "SafePtr.h"
#include "Gadgets.h"

// a basic (abstract) message event, parent for all other MsgEvents
template<class Callee>
class BasicMsgEvent: public Event {
	public:
		BasicMsgEvent(Callee *aCallee, const char *aMethodName):
			theCallee(aCallee), theMethodName(aMethodName) {}

		virtual void execute() { if (theCallee) call(); else cancel(); }

		virtual ostream &print(ostream &os) const { return printParam(os << theCallee << "->" << theMethodName << '(') << ')'; }

	protected:
		virtual void call() = 0;
		virtual ostream &printParam(ostream &os) const { return os << "..."; }

	protected:
		SmartPtr<Callee> theCallee;
		const char *theMethodName;
};

// a message event: a method call with a single [simple] parameter
template<class Callee, class Param>
class MsgEvent: public BasicMsgEvent<Callee> {
	typedef void (Callee::*Method)(Param param);

	public:
		MsgEvent(Callee *aCallee, Method aMethod, const char *aMethodName, Param aParam):
			BasicMsgEvent<Callee>(aCallee, aMethodName), theMethod(aMethod), theParam(aParam) {}

	protected:
		virtual void call() { ((*theCallee).*theMethod)(theParam); }
		virtual ostream &printParam(ostream &os) const { return os << theParam; }

	protected:
		Method theMethod;
		Param theParam;
};

// same as MsgEvent, but parameter is passed by reference 
// (perhaps the object is big and we want to copy it only once)
template<class Callee, class Param>
class ObjMsgEvent: public BasicMsgEvent<Callee> {
	typedef void (Callee::*Method)(const Param &param);

	public:
		ObjMsgEvent(Callee *aCallee, Method aMethod, const char *aMethodName, const Param &aParam):
			BasicMsgEvent<Callee>(aCallee, aMethodName), theMethod(aMethod), theParam(aParam) {}

	protected:
		virtual void call() { ((*theCallee).*theMethod)(theParam); }
		virtual ostream &printParam(ostream &os) const { return os << theParam; }

	protected:
		Method theMethod;
		Param theParam;
};

// same as MsgEvent, but the pointer-parameter must be protected
template<class Callee, class Param>
class PtrMsgEvent: public BasicMsgEvent<Callee> {
	typedef void (Callee::*Method)(Param *param);

	public:
		PtrMsgEvent(Callee *aCallee, Method aMethod, const char *aMethodName, Param *aParam):
			BasicMsgEvent<Callee>(aCallee, aMethodName), theMethod(aMethod), theParam(aParam) {}

	protected:
		virtual void call() { ((*theCallee).*theMethod)(theParam); }
		virtual ostream &printParam(ostream &os) const { return os << theParam; }

	protected:
		Method theMethod;
		SmartPtr<Param> theParam;
};

// same as MsgEvent but there is no "message", just a "ring"
template <class Callee>
class RingEvent: public BasicMsgEvent<Callee> {
	typedef void (Callee::*Method)();

	public:
		RingEvent(Callee *aCallee, Method aMethod, const char *aMethodName):
			BasicMsgEvent<Callee>(aCallee, aMethodName), theMethod(aMethod) {}

	protected:
		virtual void call() { ((*theCallee).*theMethod)(); }
		virtual ostream &printParam(ostream &os) const { return os; }

	protected:
		Method theMethod;
};


// special child->parent event
// used for child-parent communication when there is a guarantee
// that parent does not die before child (but may be ganged!)
template<class Callee, class Param>
class ChildMsgEvent: public Event {
	typedef void (Callee::*Method)(Param *param);

	public:
		ChildMsgEvent(Callee *aCallee, Method aMethod, const char *aMethodName, Param *aParam):
			theCallee(aCallee), theMethodName(aMethodName), theMethod(aMethod), theParam(aParam) {}

	protected:
		virtual void execute() { ((*theCallee).*theMethod)(theParam); }
		virtual ostream &print(ostream &os) const { return os << theCallee << "->" << theMethodName << '(' << theParam << ')'; }

	protected:
		Callee *theCallee;
		const char *theMethodName;
		Method theMethod;
		Param *theParam;
};

#endif
