// $Id: tunnel.cc,v 1.4.2.8 1998/11/24 20:46:52 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include "IOBuffer.h"
#include "Time.h"
#include "Stack.h"
#include "Socket.h"
#include "PortMgr.h"
#include "Gadgets.h"

#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h> /* for strerror */

#include <arpa/inet.h> /* for inet_ntoa */

#include <iostream.h>
#include <iomanip.h>


/*
 * Simple null-proxy
 * Tunnels traffic on a given towards given address
 */


// info for one direction of the IO stream (used by Side)
class DirRec {
	public:
		DirRec(): theBuf(0), isReady(false), isDone(true) {}

		bool done() const { return isDone; }
		bool ready() const { return isReady; }

		IOBuffer *buf() { return theBuf; }
		void buf(IOBuffer *aBuf) { theBuf = aBuf; isDone = isReady = false; }

		void beDone() { isDone = true; isReady = false; }
		void beReady(bool be = true) { isReady = be; }

	protected:
		IOBuffer *theBuf;
		bool isReady;     // fd is ready
		bool isDone;      // there should be no more IOs
};

// client or server side
class Side {
	public:
		Side(): theOtherSide(0), theFD(-1), thePort(-1), theEodStr(0), theEodBuf(0), isOpen(false), isConnected(false) { theEodBuf = new char[4096]; }
		~Side() { delete[] theEodBuf; }

		bool open() const { return isOpen; }
		bool alone() const { return !theOtherSide; }
		bool readReady() const { return theRead.ready(); }
		bool writeReady() const { return theWrite.ready(); }
		IOBuffer *readBuf() { return theRead.buf(); }
		IOBuffer *writeBuf() { return theWrite.buf(); }

		void fd(int aFD) { theFD = aFD; }
		void port(int aPort) { thePort = aPort; }
		void bufs(IOBuffer *aReadBuf, IOBuffer *aWriteBuf, const char *anEodStr = 0) { theRead.buf(aReadBuf); theWrite.buf(aWriteBuf); theEodStr = anEodStr; isOpen = true; isConnected = false; }
		void otherSide(Side *anotherSide) { theOtherSide = anotherSide; }

		void read();
		void write();
		void close();
		void stop();

	protected:
		void doneAndCheck(DirRec &toBeDone, DirRec &toBeChecked);

	protected:
		DirRec theRead;
		DirRec theWrite;

		Side *theOtherSide;
		int theFD;
		int thePort; // server side;

		const char *theEodStr; // string that indicates end-of-data on read
		char *theEodBuf;       // buffer to accumulate requests (XXX: kludge)

		bool isOpen;
		bool isConnected;
};

// wrapper around fd_set
class FD_Set {
	public:
		FD_Set();

		int maxFD() const { return theMaxFD; }
		bool active() const { return theMaxFD >= 0; }

		bool isSetFD(int fd) const { return FD_ISSET(fd, &theSet); }
		bool isReadyFD(int fd) const { return FD_ISSET(fd, &theReadySet); }

		void setFD(int fd) { FD_SET(fd, &theSet); theCount++; if (fd > theMaxFD) theMaxFD = fd; }
		void clearFD(int fd) { if (isSetFD(fd)) { FD_CLR(fd, &theSet); FD_CLR(fd, &theReadySet); theCount--; if (fd == theMaxFD) while (--theMaxFD >= 0 && !isSetFD(theMaxFD)); } }

		fd_set *readySet() { if (!active()) return 0; memcpy(&theReadySet, &theSet, sizeof(theReadySet)); return &theReadySet; }

	protected:
		fd_set theSet;
		fd_set theReadySet;

		int theMaxFD;
		int theCount;
};

typedef enum IODir { ioRead = 0, ioWrite = 1 };

// wrapper around select(2)
class Select {
	typedef Stack<IOBuffer*> BufStack;
	public:
		Select();
		~Select();

		void limits(int aFDLimit, Size bufSize);

		int count() const { return theCount; }
		FD_Set &readSet() { return theSet[ioRead]; }
		FD_Set &writeSet() { return theSet[ioWrite]; }

		void addPair(int fd1, int fd2, int port);

		bool step(); // select and fill/empty buffers

		void addFD(int fd);
		void delFD(int fd);

	protected:
		Side *theSides;
		BufStack theBufs;

		FD_Set theSet[2];

		int theFDLimit;
		int theMaxFD;
		int theCount;
};

static void Accept(Socket ls);


// globals
static const char *TheDestHost = 0;
static int ThePort = -1;
static int TheDestPort = -1;
static Select TheSelect;
static Socket TheListSock;

static int TheExpAccCount = 0;
static int TheCcLevel[2] = { 0, 0 };

// #define DEBUG_(fd) (fd <= 20)
#define DEBUG_(fd) (0)
#define DEBUG      DEBUG_(theFD)

// kludge to avoid inclusion of real Log files
ostream &TheSysLog = cerr;
ostream &TheNotepad = cout;
ostream &TheJournal = clog;
ostream &TheReport = cout;
ostream& theSysCtx(ostream& os) { return os; }
ostream& theJorCtx(ostream& os) { return os; }
ostream& thePadCtx(ostream& os) { return os; }

/* Side */

void Side::read() {
	Assert(!theRead.done());
	IOBuffer &buf = *theRead.buf();
	if (!isConnected && !theEodStr) {
		const Socket s(theFD);
		const int e = s.errNo();
		if (e)
			clog << "connect read-error?: (" << e << ") " << strerror(e) << endl;
	}
	buf.pack();
	const Size max_sz = buf.bufSize();
	if (max_sz > 0) {
		const ssize_t sz = ::read(theFD, buf.buf(), max_sz);
		buf.err(sz < 0 ? errno : 0);
		buf.eod(sz <= 0);
		if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " read: " << sz << " from " << Socket(theFD).lport() << (theRead.ready() ? " (auto)" : " (select)") << endl;
		if (sz > 0) {
			// XXX: kludge to stop reading early for HTTP requests
			//cerr << "content: " << buf.content() << endl;
			if (theEodStr) {
				Assert(buf.inOffset()+sz < 4096);
				strncpy(theEodBuf+buf.inOffset(), buf.buf(), sz);
				if (strstr(theEodBuf, theEodStr)) {
					buf.eod(true);
					doneAndCheck(theRead, theWrite);
				}
			}
			buf.appended(sz);

			// KLUDGE
			static int readCount = 10;
			if (readCount-- > 0)
				cout << "# read[" << readCount << "] " << sz << " bytes from " << Socket(theFD).peerAddr() << endl;
		} else 
		if (sz == 0) {
			buf.eod(true);
			doneAndCheck(theRead, theWrite);
			if (theOtherSide && buf.exhausted())
				theOtherSide->doneAndCheck(theOtherSide->theWrite, theOtherSide->theRead);
		} else { // (sz < 0)
			clog << "fd: " << theFD 
				<< (theEodStr ? " client-side" : " server-side")
				<< " read error: (" << buf.errNo() << ") " << strerror(buf.errNo())
				<< "; read: " << buf.inOffset() << "/" << buf.outOffset()
				<< " wrote: " << theWrite.buf()->outOffset() << "/" << theWrite.buf()->inOffset()
				<< endl;
			buf.eod(true);
			stop();
			return;
		}

		if (theOtherSide && theOtherSide->writeReady())
			theOtherSide->write();
		if (theRead.ready()) { // were not called from select
			TheSelect.readSet().setFD(theFD); // re-register
			theRead.beReady(false);
		}
	} else 
	if (!theRead.ready()) { // were called from select
		theRead.beReady(true);
		TheSelect.readSet().clearFD(theFD);
	}
	if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " after read: r: " << (theRead.done() ? "done" : theRead.ready() ? "ready" : "select") << endl;
	if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " after read: w: " << (theWrite.done() ? "done" : theWrite.ready() ? "ready" : "select") << endl;
}

void Side::write() {
	Assert(!theWrite.done());
	IOBuffer &buf = *theWrite.buf();
	const Size max_sz = buf.contSize();
	if (!isConnected && !theEodStr) {
		Assert(!theWrite.ready());
		const Socket s(theFD);
		const int e = s.errNo();
		if (e) {
			const timeval t = s.rcvTimeout();
			clog << "connect error?: (" << e << ") " << strerror(e) << endl;
			clog << "fd: " << theFD << " rcv-tout: " << t.tv_sec << ':' << t.tv_usec << endl;
			buf.err(e);
            stop();
			return;
		}
		isConnected = true;
	}
	if (max_sz > 0) {
		const ssize_t sz = ::write(theFD, buf.content(), max_sz);
		buf.err(sz < 0 ? errno : 0);
		if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " wrote: " << sz << " to " << Socket(theFD).lport() << (theWrite.ready() ? " (auto)" : " (select)") << endl;
		if (sz > 0) {
			buf.consumed(sz);

			// KLUDGE
			static int writeCount = 10;
			if (writeCount-- > 0)
				cout << "# wrote[" << writeCount << "] " << sz << " bytes to " << Socket(theFD).peerAddr() << endl;
		} else
		if (sz < 0) {
			clog << "fd: " << theFD 
				<< (theEodStr ? " client-side" : " server-side")
				<< " write error: (" << buf.errNo() << ") " << strerror(buf.errNo())
				<< endl;
			stop();
			return;
		}
		if (buf.exhausted())
			doneAndCheck(theWrite, theRead);
		if (theOtherSide && theOtherSide->readReady())
			theOtherSide->read();
		if (theWrite.ready()) { // were not called from select
			TheSelect.writeSet().setFD(theFD); // re-register
			theWrite.beReady(false);
		}
	} else 
	if (!theWrite.ready()) { // were called from select
		theWrite.beReady(true);
		TheSelect.writeSet().clearFD(theFD);
	}
	if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " after write: w: " << (theWrite.done() ? "done" : theWrite.ready() ? "ready" : "select") << endl;
	if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " after write: r: " << (theRead.done() ? "done" : theRead.ready() ? "ready" : "select") << endl;
}

void Side::close() {
	if (DEBUG) cerr << Time::Now() << " fd: " << theFD << " closed" << endl;
	TheSelect.delFD(theFD); // will close
	if (thePort >= 0)
		ThePortMgr.free(thePort, !theWrite.buf()->err() && !theRead.buf()->err());
	isConnected = isOpen = false;
	theRead.buf(0);
	theWrite.buf(0);
	if (theOtherSide)
		theOtherSide->otherSide(0);
	if (theEodStr) // XXX: ugly (we are the client side)
		TheExpAccCount++;
	TheCcLevel[!theEodStr]--;
}

void Side::doneAndCheck(DirRec &toBeDone, DirRec &toBeChecked) {
	if (!toBeDone.ready()) // may have registration
		(&toBeDone == &theRead) ? // XXX: ugly
			TheSelect.readSet().clearFD(theFD) :
			TheSelect.writeSet().clearFD(theFD);
	toBeDone.beDone();
	if (toBeChecked.done())
		close();
}

void Side::stop() {
	// may have registrations
	if (!theRead.ready())
		TheSelect.readSet().clearFD(theFD);
	if (!theWrite.ready())
		TheSelect.writeSet().clearFD(theFD);
	theRead.beDone();
	theWrite.beDone();
	close();
	if (theOtherSide)
		theOtherSide->stop();
	theOtherSide = 0;
}

/* FD_Set */

FD_Set::FD_Set(): theMaxFD(-1), theCount(0) {
	FD_ZERO(&theSet);
	FD_ZERO(&theReadySet);
}

/* Select */

Select::Select(): theSides(0), theFDLimit(-1), theMaxFD(-1), theCount(0) {
}

Select::~Select() {
	delete[] theSides;
	while (theBufs.count()) delete theBufs.pop();
}

void Select::limits(int aFDLimit, Size bufSize) {
	Assert(theFDLimit < 0);
	theFDLimit = aFDLimit;
	theSides = new Side[theFDLimit];
	theBufs.stretch(theFDLimit);
	for (int i = 0; i < theFDLimit; ++i)
		theSides[i].fd(i);
	cerr << "allocated " << ((sizeof(Side)*theFDLimit)/1024.) << " KB of side-objects for " << theFDLimit << " sockets" << endl;
	cerr << "allocated " << (4096*theFDLimit)/1024. << " KB of side-objects EOD--buffers for " << theFDLimit << " sides" << endl;
	Size sum = 0;
	for (int i = 0; i < theBufs.capacity(); ++i) {
		IOBuffer *buf = new IOBuffer();
		buf->minCapacity(bufSize);
		theBufs.push(buf);
		sum += buf->bufSize();
	}
	cerr << "allocated " << sum/1024. << " KB of buffers for about " << theBufs.capacity()-1 << " connections" << endl;
}

void Select::addPair(int fd1, int fd2, int port) {
	addFD(fd1);
	addFD(fd2);
	if (DEBUG) cerr << Time::Now() << " pair: fd: " << fd1 << " and " << fd2 << " port: " << port << endl;
	Assert(theBufs.count() >= 2);
	IOBuffer *b1 = theBufs.pop(); b1->reset();
	IOBuffer *b2 = theBufs.pop(); b2->reset();
	theSides[fd1].bufs(b1, b2, "\n\r\n");
	theSides[fd2].bufs(b2, b1);
	theSides[fd1].otherSide(theSides+fd2);
	theSides[fd2].otherSide(theSides+fd1);
	theSides[fd1].port(-1);
	theSides[fd2].port(port);
	TheCcLevel[0]++, TheCcLevel[1]++;
}

void Select::addFD(int fd) {
	Assert(fd >= 0 && fd <= theFDLimit);
	Assert(!theSides[fd].open());
	Assert(!theSet[ioRead].isSetFD(fd) && !theSet[ioWrite].isSetFD(fd));
	theSet[ioRead].setFD(fd);
	if (fd != TheListSock) // kludge
		theSet[ioWrite].setFD(fd);
	if (fd > theMaxFD)
		theMaxFD = fd;
	theCount++;
	if (DEBUG_(fd)) cerr << "fd: " << fd << " opened; max: " << theMaxFD << "; count: " << theCount << endl;
}

void Select::delFD(int fd) {
	Assert(fd >= 0 && fd <= theFDLimit);
	Assert(fd == TheListSock || theSides[fd].open());
	theSet[ioRead].clearFD(fd);
	theSet[ioWrite].clearFD(fd);
	::close(fd);
	if (fd == theMaxFD)
		while (--theMaxFD >= 0 && theMaxFD != TheListSock && !theSides[theMaxFD].open());
	if (fd != TheListSock && theSides[fd].alone()) {
		theBufs.push(theSides[fd].readBuf());
		theBufs.push(theSides[fd].writeBuf());
	}
	theCount--;
		
	//cerr << "fd: " << fd << " closed; max: " << theMaxFD << "; count: " << theCount << endl;
}

bool Select::step() {
	struct timeval tout = { theCount*60, 0 };
	Assert(theCount);
	Assert(theMaxFD >= 0);

	if (DEBUG) cerr << Time::Now() << " select(" << theMaxFD << "/" << theCount << ")" << endl;
	Assert(theSet[ioRead].isSetFD(TheListSock));

	const int count = ::select(theMaxFD+1, theSet[ioRead].readySet(), theSet[ioWrite].readySet(), 0, &tout);
	if (count < 0) {
		Check(errno == EINTR);
		return true;
	}
	if (count == 0) {
		cerr << Time::Now() << " exiting after " << tout.tv_sec << " seconds of inactivity" << endl;
		return false;
	}
		
	// accept
	if (theSet[ioRead].isReadyFD(TheListSock))
		Accept(TheListSock);

	// reads
	for (int fd = theSet[ioRead].maxFD(); fd >= 0; --fd) {
		if (theSet[ioRead].isReadyFD(fd) && fd != TheListSock) {
			theSides[fd].read();
		}
	}

	// accept
	if (theSet[ioRead].isReadyFD(TheListSock))
		Accept(TheListSock);

	// writes
	for (int fd = theSet[ioWrite].maxFD(); fd >= 0; --fd) {
		if (theSet[ioWrite].isReadyFD(fd)) {
			theSides[fd].write();
		}
	}

	// accept
	if (theSet[ioRead].isReadyFD(TheListSock))
		Accept(TheListSock);

	return true;
}

static
Socket CreateSock() {
	Socket s;
	Check(s.create());
	s.blocking(false);
	s.linger(0);
	//s.reuseAddr(true);
	return s;
}

static
void Accept(Socket ls) {
	static Time lastCall = Time::Now();
	static Time delayRecord;
	static int accRecord;
	static int ccRecord[2] = { 0, 0};
	static int callCount = 0;

	callCount++;
	//if (callCount % 1000 == 0)
	//	cerr << "Accept[" << callCount << "]: cc level: " << TheCcLevel[0] << '/' << TheCcLevel[1] << endl;

	if (TheCcLevel[0] > ccRecord[0]) {
		cerr << "Accept[" << callCount << "]: record cc[0] level: " << TheCcLevel[0] << '/' << TheCcLevel[1] << endl;
		ccRecord[0] = TheCcLevel[0];
	}
	if (TheCcLevel[1] > ccRecord[1]) {
		cerr << "Accept[" << callCount << "]: record cc[0] level: " << TheCcLevel[0] << '/' << TheCcLevel[1] << endl;
		ccRecord[1] = TheCcLevel[1];
	}

	const Time t = Time::Now();
	const Time d = t - lastCall;
	if (d > delayRecord) {
		cerr << "Accept[" << callCount << "]: record delay of " << d << " expect: " << TheExpAccCount << endl;
		delayRecord = d;
	}

	const int expCount = TheExpAccCount;
	int count = 0;
	do {
		Socket as = ls.accept(0, 0);
		if (as < 0) {
			Check(errno == EWOULDBLOCK);
			break;
		}
		count++;
		lastCall = t;

		// ThePortMgr.reserve(as.port());
		Socket cs = CreateSock();
		int attemptCount = 0;
		// we have to connect because we do not want to re-queue or 
		// drop accepted socket
		int port;
		while ((port = ThePortMgr.bind(cs)) < 0 || !cs.connect(TheDestHost, TheDestPort)) {

			static MaxAttemptCount = 1;
			if (attemptCount > MaxAttemptCount) {
				MaxAttemptCount = attemptCount;
				clog << " connect-attempt record: " << MaxAttemptCount << endl;
			}

			if (++attemptCount > 1000) {
				cerr << " failed to connect; tried " << attemptCount << " times" << endl;
				exit(-2);
			}

			cs.close();
			cs = CreateSock();
		}
		TheSelect.addPair(as, cs, port);
	} while (1); // --TheExpAccCount > 0);
	TheExpAccCount = 0;

	if (count > accRecord) {
		cerr << "Accept[" << callCount << "]: record accepted count of " << count << " exp: " << expCount << endl;
		accRecord = count;
	}
	if (DEBUG) cerr << Time::Now() << " Accept[" << callCount << "]: accepted: " << count << " exp: " << expCount << endl;
}

int main(int argc, char *argv[]) {
	char *h;
	if (argc < 3 || argc > 4 || !(h = strchr(argv[2], ':'))) {
		cerr << "usage: " << argv[0] << " <listen_port> <dest_host:dest_port> [port_range] " << endl;
		return -1;
	}

	// ignore stupid signals
	signal(SIGPIPE, SIG_IGN);

	ThePort = atoi(argv[1]);
	*h = '\0';
	TheDestHost = argv[2];
	TheDestPort = atoi(h+1);

	cerr << "connecting localhost:" << ThePort << " and " << TheDestHost << ':' << TheDestPort << endl;

	if (argc == 4) {
		h = strchr(argv[3], ':');
		const int min = xatoi(argv[3], -1);
		const int max = xatoi(h + 1, -1);
		if (max >= min) {
			ThePortMgr.configure(min, max, max >= min);
			cerr << "managing ports from " << min << " to " << max << endl;
		} else
			cerr << "using ephemeral ports" << endl;
	} else
		cerr << "using ephemeral ports" << endl;

	ThePortMgr.reserve(ThePort);

	TheListSock = CreateSock();
	TheListSock.reuseAddr(true);
	//TheListSock.reusePort(true);
	Check(TheListSock.listen(ThePort));

	SetMaxFD(GetMaxFD());
	TheSelect.limits(GetMaxFD()+1, Size::KB(4)-1);
	TheSelect.addFD(TheListSock);
	
	while (TheSelect.step()) {};

	TheSelect.delFD(TheListSock);
	return 0;
}
